#lang racket

(require "main.rkt" (except-in "cps.rkt" lookup F ∪⊗)
         "util.rkt"
         redex
         srfi/1/list)

;; Lemma 4.1
(define (l4.1)
  (redex-check anf
               (t e)
               (equal? (term ((μ t e)))
                       (term (μc t (e))))
               #:source μ))

;; just mentioned in running text that F = Fc
(define (sec4-para3)
  (redex-check anf
               p
               (equal? ((F (term p)) '())
                       ((Fc (term p)) '()))))


(define (iff a b) (equal? a b))

(define (triple< t1 t2)
  (and (lset<= equal? (first t1) (first t2))
       (lset<= equal? (second t1) (second t2))
       (lset<= equal? (third t1) (third t2))))

(define (set< t1 t2) (lset<= equal? t1 t2))

(define-syntax-rule
  (galois left right comp1 comp2 alpha gamma)
  (redex-check anf
               (left right)
               (iff (comp1 (term (alpha left)) (term right))
                    (comp2 (term left) (term (gamma right))))))

;; Lemma 4.2
(define (l4.2)
  (galois (M ...) ((C ...) ([C K] ...) (Env ...))
          triple< set<
          αx γx)
  (redex-check anf 
               ((C ...) ([C K] ...) (Env ...) (M ...))
               (iff (triple< (term (αx (M ...))) (term ((C ...) ([C K] ...) (Env ...))))
                    (set< (term (M ...)) (term (γx ((C ...) ([C K] ...) (Env ...))))))))

;; Theorem 4.1
(define (t4.1)
  (redex-check anf
               (p (C_1 ...) ([C_2 K] ...) (Env ...))
               (equal? ((Fx (term p)) (term ((C_1 ...) ([C_2 K] ...) (Env ...))))
                       (term (αx ,((Fc (term p)) (term (γx ((C_1 ...) ([C_2 K] ...) (Env ...))))))))))


;; connect the abstract interpretation to evaluation
(define (c4.1-ish)
  (redex-check 
   anf
   s
   (let* ([s* (bind-fv (term s))]
          [analysis-result (F#-answers (lfp (F# s*) '(() () ())))]
          [machine-result (with-handlers ((exn:fail? (λ (x) 'error)))
                            (match (Eval s*)
                              [`(close ,fn ,env) fn]
                              [w w]))])
     (equal? (or (eq? machine-result 'error)
                 (and (member machine-result analysis-result) #t))
             #t))))

(define (cps-sound)
  (redex-check 
   cps
   p
   (let* ([s* (bind-fv (term p))]
          [analysis-result (T#-answers (lfp (T# (term (C ,s*))) '(() ())))]
          [machine-result (with-handlers ((exn:fail? (λ (x) 'error)))
                            (match (Eval s*)
                              [`(close ,fn ,env) fn]
                              [w w]))])
     (equal? (or (eq? machine-result 'error)
                 (and (member machine-result analysis-result) #t))
             #t))))


(define (go)
  (sec4-para3)
  ;(l4.1) ;; cannot generate all of the lhs's 
  (l4.2)
  (t4.1)
  (c4.1-ish)
  (cps-sound))
  
(define (t5.1)  
  (define-extended-language restricted-vars cps
    (fresh-non-terminal xr vr kr))
  (redex-check restricted-vars p 
               (let ([p* (bind-fv (term p))])
                 (with-handlers ([exn:fail? void])
                   (equal? (T#-answers (lfp (T# (term (C ,p*))) '(() ())))
                           (F#-answers (lfp (F# p*) '(() () ()))))))))

(define (t5.1/no-c [n 1000])
  (define-extended-language restricted-vars cps
    (fresh-non-terminal xr vr kr))  
  (redex-check restricted-vars nce-s 
               (let ([p* (bind-fv-id (term nce-s))])
                 (with-handlers ([exn:fail? void])
                   (equal? (T#-answers (lfp (T# (term (C ,p*))) '(() ())))
                           (F#-answers (lfp (F# p*) '(() () ()))))))
               #:attempts n))
#;
(go)