#lang scheme

(require redex/reduction-semantics
         "test-common.rkt"
         "reduction.rkt"
         "typing.rkt")

(test-equal (term (subst (inc x) x 7))
            (term (inc 7)))
(test-equal (term (subst (block (def f (x : (Int ε)) : (Int ε) = x) f) x 7))
            (term (block (def f (x : (Int ε)) : (Int ε) = x) f)))
(test-equal (term (subst (block (def f (x : (Int ε)) : (Int ε) = x) f) f 7))
            (term (block (def f (x : (Int ε)) : (Int ε) = x) f)))
(test-equal (term (subst (block (def f (x : (Int ε)) : (Int ε) = y) y) y 7))
            (term (block (def f (x : (Int ε)) : (Int ε) = 7) 7)))
(test-equal (term (subst (block (def f (y : (Int ε)) : (Int ε) = y) y) y 7))
            (term (block (def f (y : (Int ε)) : (Int ε) = y) 7)))
(test-equal (term (subst (block (def f (z : (Int ε)) : (Int ε) = (f z)) x) x y))
            (term (block (def f1 (z1 : (Int ε)) : (Int ε) = (f1 z1)) y)))
(test-equal (term (subst (block (def f (z : (Int ε)) : (Int ε) = (block f z x)) f z x) x f))
            (term (block (def f1 (z1 : (Int ε)) : (Int ε) = (block f1 z1 f)) f1 z f)))
(test-equal (term (subst (block (val y : (Int ε) = x) y x) x y))
            (term (block (val y1 : (Int ε) = y) y1 y)))
(test-equal (term (subst (block (val x : (Int ε) = x) x) x z))
            (term (block (val x : (Int ε) = z) x)))

; CBV application
(let ([t (term ((block (def f (x : (Int ε)) : (Int ε) = (f x)) f) 1))])
  (test--> ->sr t t))
(test--> ->sr (term ((block (def f (f : (Int ε)) : (Int ε) = f) f) 1)) 1)

; Primitives
(test--> ->sr (term (dbl (inc 0))) (term (dbl 1)))
(test--> ->sr (term (dbl 1)) 2)
(test--> ->sr (term (len "abc")) 3)

; Block
(test--> ->sr (term (block 1 (inc 2))) (term (block (inc 2))))
(test--> ->sr (term (block 1 (val x : (Int ε) = 7) x)) (term (block (val x : (Int ε) = 7) x)))
(test--> ->sr (term (block (inc 2))) (term (inc 2)))
(test--> ->sr (term (block (inc 0) (inc 1))) (term (block 1 (inc 1))))

; Val
(test--> ->sr (term (block (val x : (Int ε) = 0) x)) (term (block 0)))
(test--> ->sr 
         (term (block (val x : (Int ε) = (inc 0)) x))
         (term (block (val x : (Int ε) = 1) x)))
(test--> ->sr
         (term (block 
                (val x : (Int ε) = 0)
                (val x : (Int ε) = (inc x))
                x))
         (term (block 
                (val x : (Int ε) = (inc 0))
                x)))

; Def
(test--> ->sr 
         (term (block (def add2 (i : (Int ε)) : (Int ε) = (inc (inc i))) (add2 0)))
         (term (block ((block (def add2 (i : (Int ε)) : (Int ε) = (inc (inc i))) add2) 0))))
(test--> ->sr
         (term (block 
                (def x (y : (Int ε)) : (Int ε) = 7)
                (val x : (Int ε) = (x 3))
                x))
         (term (block 
                (val x : (Int ε) = ((block (def x (y : (Int ε)) : (Int ε) = 7) x) 3))
                x)))

; Shift
(test-->> ->sr shift-example-1 0)
(test-->> ->sr shift-example-2 1)
(test-->> ->sr shift-example-3 9)
(test-->> ->sr shift-example-4 2)
(test-->> ->sr shift-example-5 1)

; Reset
(test--> ->sr (term (reset 5)) 5)

(test-results)