#lang scheme

(require redex/reduction-semantics
         "test-common.rkt"
         "typing.rkt"
         "reduction.rkt")

(test-predicate well-typed? (term inc))
(test-predicate well-typed? (term dbl))
(test-predicate well-typed? (term len))
(test-predicate well-typed? (term (dbl 1)))
(test-predicate well-typed? (term (len "abc")))
(test-predicate well-typed? (term (block
                                   (def 2x+1 (x : (Int ε)) : (Int ε) =
                                     (inc (dbl x)))
                                   (2x+1 7))))
(test-predicate well-typed? (term (block 1 (def x (x : (Int ε)) : (Int ε) = x) x)))
(test-predicate well-typed? (term (block (val x : (Int ε) = 7) (inc x))))
(test-predicate well-typed? (term (block 1 2 3)))
(test-predicate 
 well-typed?
 (term (reset (inc (shift (block (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Str ε) = "s") f))))))

(let ([t (term 
          (block
           (def f (s : (Str ε)) : (Str ε) = s)
           (def g (k : ((⇒ (Str ε) (Int ε)) ε)) : (Int ε) = (dbl (k "s")))
           (f (shift g))))])
  (test-equal (term (typeof () ,t)) (term (Str (@cps Int Int))))
  (test-predicate (negate well-typed?) t)
  (test-predicate well-typed? (term (reset (len ,t)))))

(term-let 
  ([i (term (Int ε))])
  (term-let 
   ([i->i (term ((⇒ i i) ε))])
   (term-let
    ([i->i->i (term ((⇒ i i->i) ε))])
    (term-let
     ([i->i->i->i (term ((⇒ i i->i->i) ε))])
     (test-predicate
      well-typed?
      (term 
       (block
        (def t1 (x : i) : i = x)
        (def t2 (_ : i) : i->i = t1)
        (def t3 (_ : i) : i->i->i = t2)
        (def f (k : i->i) : i->i = k)
        (def g (x : i) : (Int (@cps Int (⇒ i i))) = (shift f))
        (def h (k : ((⇒ ((⇒ i (Int (@cps Int (⇒ i i)))) ε) i->i->i) ε)) : i->i->i->i = 
          (block (def q (_ : i) : i->i->i = (k g)) q))
        (def j (k : ((⇒ i i->i) ε)) : i->i->i = 
          (block (def q (x : i) : i->i = (k x)) q))
        (reset (inc ((shift h) (shift j)))))))
     (test-predicate
      well-typed?
      (term 
       (block
        (def f (k : ((⇒ i i->i) ε)) : i = ((k 1) 2))
        (def g (k : ((⇒ i i->i->i) ε)) : i->i = ((k 3) 4))
        (def h (k : ((⇒ i i->i->i->i) ε)) : i->i->i = ((k 5) 6))
        (def j (x : i) : i->i->i->i =
          (block
           (def t1 (_ : i) : i->i->i = 
             (block
              (def t2 (_ : i) : i->i = 
                (block (def t3 (_ : i) : i = x) t3))
              t2))
           t1))
        (reset (j (block (shift f) (shift g) (shift h)))))))
     
     (test-predicate
      well-typed?
      (term 
       (block
        (def f (k : ((⇒ i i->i) ε)) : i->i->i = 
          (block (def q (x : i) : i->i = (k x)) q))
        (def g (k : ((⇒ i i) ε)) : i->i = k)
        (reset (block (val x : i = (shift f)) (shift g))))))))))

(test-predicate well-typed? shift-example-1)
(test-predicate well-typed? shift-example-2)
(test-predicate well-typed? shift-example-3)
(test-predicate well-typed? shift-example-4)
(test-predicate well-typed? shift-example-5)

(test-predicate (negate well-typed?) (term (inc "s")))
(test-predicate (negate well-typed?) (term (len 7)))
(test-predicate (negate well-typed?) (term ("s" len)))

(test-predicate
 (negate well-typed?)
 (term
  (reset
   (block
    (val x : (Int ε) = 
         (shift 
          (block
           (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (dbl (k 4)))
           f)))
    (shift 
     (block
      (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Str ε) = "s")
      f))))))

(test-predicate
 (negate well-typed?)
 (term
  (reset
   (block
    (shift 
     (block
      (def f (k : ((⇒ (Str ε) (Int ε)) ε)) : (Int ε) = (dbl (k "s")))
      f))
    (shift
     (block
      (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Str ε) = "s")
      f))))))

(test-predicate
 (negate well-typed?)
 (term
  (reset
   ((shift 
     (block
      (def f (k : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε)) : (Int ε) =
        (inc (k (block (def g (x : (Int ε)) : (Int ε) = x) g))))
      f))
    (shift 
     (block
      (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Str ε) = "s")
      f))))))

(test-predicate
 (negate well-typed?)
 (term
  (reset
   ((block
     (def f (x : (Int ε)) : (Int (@cps Int Str)) =
       (shift (block (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Str ε) = "s") g)))
     f)
    (shift 
     (block (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k 7)) f))))))

(test-predicate
 (negate well-typed?)
 (term
  (reset
   ((shift 
     (block
      (def f (k : ((⇒ ((⇒ (Int ε) (Int (@cps Int Str))) ε) (Int ε)) ε)) : (Int ε) =
        (k (block
            (def g (x : (Int ε)) : (Int (@cps Int Str)) =
              (shift (block (def h (k : ((⇒ (Int ε) (Int ε)) ε)) : (Str ε) = "s") h)))
            g)))
      f))
    7))))

(test-results)