#lang racket

(require redex
         "redex.rkt"
         "grammar.rkt"
         "eval.rkt"
         "solve.rkt"
         "inference.rkt")

(port-print-handler (current-error-port) pretty-print)

(define (test)
  (redex-test-suite
    (test-empty)
    (test-grammar)
    (test-eval)
    (test-solve)
    (test-infer)))

(test)

#|

(define-syntax-rule (time* e)
  (begin
    (collect-garbage)
    (let ([x (time (with-handlers ([exn:fail? values]) e))])
      (if (exn:fail? x) (raise x) x))))

(define-extended-language typed-gadt-lang gadt-lang
  [prog: (side-condition prog_0 (well-typed? (term prog_0)))])

(define (sound? p)
  (if (well-typed? p)
      (match (apply-reduction-relation* run-gadt p)
        [(list r)
         (and (well-typed? r)
              (equal? (type p)
                      (type r)))]
        [_ #f])
      #t))

(define (well-typed? p)
  (term (is-typed ,p)))

(define (type p)
  (term (type-of ,p)))

(define-relation gadt-lang
  [(is-typed prog) (¬ (has-type prog)) (has-type (infer prog))])

(define-relation gadt-lang
  [(has-type (program dd ... (e : τ)))
   ;; don't want to consider α-equivalence
   (≡ (fv τ) {})])

(define-metafunction gadt-lang
  [(type-of prog) (get-type (infer prog))])

(define-metafunction gadt-lang
  [(get-type (program dd ... (e : τ))) τ])

(define (test-soundness a r)
  (redex-check typed-gadt-lang prog: sound?
               #:attempts a
               #:retries r))

(define (test-speed f s a r)
  (redex-check gadt-lang prog (fast? f s)
               #:attempts a
               #:retries r))

(define ((fast? f s) p)
  (let ([t (thread (lambda () (f p)))])
    (if (sync/timeout s (thread-dead-evt t))
        #t
        (begin (kill-thread t) #f))))

(define (test-auto a r)
  (time* (test-soundness a r)))

|#
