#lang racket

(require redex "grammar.rkt" "entailment.rkt")

(provide (all-defined-out))

(define-relation gadt-lang
  [(⊢ (program dd ... e))
   (⊢ [^] {} (build-delta dd ...) e)]
  [(⊢ C Γ ∆ ((x : υ) : τ)) (⊨ C [^ (τ ~ υ)]) (⊢ C Γ ∆ (x : υ))]
  [(⊢ C Γ ∆ ((x @ φ) : τ)) (⊢ C Γ ∆ ((rename φ x) : τ))]
  [(⊢ C Γ ∆ (x : τ)) (∈ x (domain Γ)) (reconcilable (lookup Γ x) [^] τ)]
  [(⊢ C Γ ∆ (K : τ)) (∈ K (domain ∆)) (reconcilable (lookup ∆ K) C τ)]
  [(⊢ C Γ ∆ ((λ x (e : υ)) : (→ τ υ))) (⊢ C (∪ Γ {[x := τ]}) ∆ (e : υ))]
  [(⊢ C Γ ∆ (((e_1 : (→ υ τ)) (e_2 : υ)) : τ))
   (⊢ C Γ ∆ (e_1 : (→ υ τ)))
   (⊢ C Γ ∆ (e_2 : υ))]
  [(⊢ C Γ ∆ ((let (g = (e_1 : τ_1)) in (e_2 : τ_2)) : τ_2))
   (⊢ C (∪ Γ {[g : τ_1]}) ∆ (e_1 : τ_1))
   (⊢ C 
      (∪ Γ {[g := (∀ (∖ (fv τ_1) (∪ (fv C) (fv Γ))) [^] ⇒ τ_1)]})
      ∆
      (e_2 : τ_2))]
  [(⊢ C Γ ∆ ((let (g :: (∀ (a ...) [^] ⇒ τ_1) = (e_1 : τ_1))
               in
               (e_2 : τ_2)) : τ_2))
   (⊢ C (∪ Γ {[g := (∀ (a ...) [^] ⇒ τ_1)]}) ∆ (e_1 : τ_1))
   (⊢ C (∪ Γ {[g := (∀ (a ...) [^] ⇒ τ_1)]}) ∆ (e_2 : τ_2))]
  [(⊢ C Γ ∆ ((case (e_1 : τ_1) of [p → e] ...) : τ_2))
   (⊢ C Γ ∆ (e_1 : τ_1))
   (∧ (⊢p C Γ ∆ [p → e] τ_1 τ_2) ...)]
  )

(define-metafunction gadt-lang
  [(⊢p C Γ ∆ [(K x ...) →  (e : τ_r)] (T τ ...) τ_r)
   (∧ (≡ (∩ (∪ (fv C) (fv Γ) (fv τ) ... (fv τ_r)) {b ...}) {})
      (consistent (& C (rename φ D)))
      (⊢ (& C (rename φ D))
         (∪ Γ (rename φ {[x := υ] ...}))
         ∆
         (e : τ_r)))
   (side-condition (term (∈ K (domain ∆))))
   (where (∀ (c ...) D ⇒ τ_1) (lookup K ∆))
   (where ((→ υ ... (T a ...)) (a ... b ...))
          ((flatten-arrow τ_1) (c ...)))
   (side-condition (term (≈ (a ...) (τ ...))))
   (where φ {[a := τ] ...})
   ])

(define-relation gadt-lang
  [(reconcilable υ C τ) (reconcilable (∀ () [^] ⇒ υ) C τ)]
  [(reconcilable (∀ (a ...) D ⇒ υ) C τ)
   (≡ τ (rename (instantiation (a ...) υ τ) υ))
   (⊨ C (rename (instantiation (a ...) υ τ) D))])

(define-metafunction gadt-lang
  [(instantiation (a ... b c ...) b τ) {[b := τ]}]
  [(instantiation (a ...) ν ν) {}]
  [(instantiation (a ...) (→ υ_1 υ_2) (→ τ_1 τ_2))
   (∪ (instantiation (a ...) υ_1 τ_1)
      (instantiation (a ...) υ_2 τ_2))]
  [(instantiation (a ...) (T υ ...) (T τ ...))
   (∪ (instantiation (a ...) υ τ) ...)
   (side-condition (term (≈ (υ ...) (τ ...))))]
  [(instantiation (a ...) υ τ) {}])
